/*
 *  ATSHELL - OS/2 RUNWORKPLACE replacement.
 *  Copyright (C) 2005 Alex Taylor
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "atshell.h"

/* ------------------------------------------------------------------------- *
 * Main program loop.                                                        *
 * ------------------------------------------------------------------------- */
int main( int argc, char *argv[] )
{
    HAB      hab;                      // anchor block handle
    HMQ      hmq;                      // message queue handle
    HWND     hwndFrame,                // window handle
             hwndAccel,                // acceleration table
             hwndHelp;                 // help instance
    QMSG     qmsg;                     // message queue
    BOOL     fInitFailure = FALSE,
             fUserQuit    = FALSE;
    UCHAR    szModule[ 16 ] = "",
             szError[ 256 ] = "Unknown error";
    ULONG    ulrc;
    USHORT   i;
    HMODULE  hResources;               // resource module handle
    HELPINIT helpInit;                 // help init structure

    ATSHGLOBAL global;


    hab = WinInitialize( 0 );
    if ( hab == NULLHANDLE ) {
        sprintf( szError, "WinInitialize() failed.");
        fInitFailure = TRUE;
    }

    if ( ! fInitFailure ) {
        hmq = WinCreateMsgQueue( hab, 0 );
        if ( hmq == NULLHANDLE ) {
            sprintf( szError, "Unable to create message queue:\nWinGetLastError() = 0x%X\n", WinGetLastError(hab) );
            fInitFailure = TRUE;
        }
    }

    if ( ! fInitFailure ) {
        ulrc = DosLoadModule( (PSZ) szModule, sizeof(szModule), RESOURCES, &hResources );
        if ( ulrc != NO_ERROR ) {
            sprintf( szError, "Failed to load resource library %s:\nDosLoadModule() = %d (\"%s\")", RESOURCES, ulrc, szModule );
            fInitFailure = TRUE;
        }
    }

    if ( ! fInitFailure ) {

        global.hab        = hab;
        global.hmq        = hmq;
        global.hResources = hResources;
        global.fRunSilent = FALSE;
        global.fRunSecure = FALSE;

        // Initialize the global data
        global.hIni       = PrfOpenProfile( hab, LocateProfile() );

        for ( i = 1; i < argc; i++ ) {
            if ( strnicmp( argv[ i ], "/q", 2 ) == 0 ) global.fRunSilent = TRUE;
            if ( strnicmp( argv[ i ], "/s", 2 ) == 0 ) global.fRunSecure = TRUE;
        }

        global.pszProgram  = NULL;
        global.pszParams   = NULL;
        global.pszDir      = NULL;
        global.pszEnv      = NULL;
        global.pszSProgram = NULL;
        global.pszSParams  = NULL;
        global.pszSDir     = NULL;
        global.pszSEnv     = NULL;

        // Read settings from the profile (INI) file
        LoadProfileSetting( &(global.pszProgram),  global.hIni, PRF_DEFAULT_PROGRAM, PRF_KEY_PROGRAM );
        LoadProfileSetting( &(global.pszParams),   global.hIni, PRF_DEFAULT_PROGRAM, PRF_KEY_PARAMS );
        LoadProfileSetting( &(global.pszDir),      global.hIni, PRF_DEFAULT_PROGRAM, PRF_KEY_DIR );
        LoadProfileSetting( &(global.pszEnv),      global.hIni, PRF_DEFAULT_PROGRAM, PRF_KEY_ENV );
        LoadProfileSetting( &(global.pszSProgram), global.hIni, PRF_STARTUP_PROGRAM, PRF_KEY_PROGRAM );
        LoadProfileSetting( &(global.pszSParams),  global.hIni, PRF_STARTUP_PROGRAM, PRF_KEY_PARAMS );
        LoadProfileSetting( &(global.pszSDir),     global.hIni, PRF_STARTUP_PROGRAM, PRF_KEY_DIR );
        LoadProfileSetting( &(global.pszSEnv),     global.hIni, PRF_STARTUP_PROGRAM, PRF_KEY_ENV );
        if ( global.pszProgram == NULL ) global.pszProgram = "PMSHELL.EXE";

        // Now load the main dialog window
        hwndFrame = WinLoadDlg( HWND_DESKTOP, HWND_DESKTOP, ClientWndProc,
                                hResources, ID_MAINPROGRAM, &global        );

        if ( hwndFrame == NULLHANDLE ) {
            sprintf( szError, "Failed to load dialog resource:\nWinGetLastError() = 0x%X\n", WinGetLastError(hab) );
            fInitFailure = TRUE;
        }
    }

    if (( fInitFailure == TRUE ) || ( hwndFrame == NULLHANDLE ))
    {
        WinMessageBox( HWND_DESKTOP, HWND_DESKTOP, szError, "Program Initialization Error", 0, MB_CANCEL | MB_ERROR );
    } else {

        // Initialize acceleration table
        hwndAccel = WinLoadAccelTable( hab, global.hResources, ID_MAINPROGRAM );
        WinSetAccelTable( hab, hwndAccel, hwndFrame );

        // Initialize online help
        helpInit.cb                         = sizeof( HELPINIT );
        helpInit.pszTutorialName            = NULL;
        helpInit.phtHelpTable               = (PHELPTABLE) MAKELONG( ID_MAINPROGRAM, 0xFFFF );
        helpInit.hmodHelpTableModule        = global.hResources;
        helpInit.hmodAccelActionBarModule   = global.hResources;
        helpInit.fShowPanelId               = 0;
        helpInit.idAccelTable               = 0;
        helpInit.idActionBar                = 0;
        helpInit.pszHelpWindowTitle         = "Help";
        helpInit.pszHelpLibraryName         = HELP_FILE;

        hwndHelp = WinCreateHelpInstance( hab, &helpInit );
        WinAssociateHelpInstance( hwndHelp, hwndFrame );

        while ( WinGetMsg( hab, &qmsg, 0, 0, 0 )) WinDispatchMsg( hab, &qmsg );

    }
    PrfCloseProfile( global.hIni );
    WinDestroyWindow( hwndFrame );
    WinDestroyMsgQueue( hmq );
    WinTerminate( hab );

    return ( 0 );
}


/* ------------------------------------------------------------------------- *
 * ------------------------------------------------------------------------- */
MRESULT EXPENTRY ClientWndProc( HWND hwnd, ULONG msg, MPARAM mp1, MPARAM mp2 )
{
    static   PATSHGLOBAL pGlobal;
    static   HWND        hwndMenu;
    static   BOOL        fRunNow = FALSE;
    HPOINTER hicon;
    POINTL   mouse;
    HWND     hwndHelp;
    BOOL     fSuccess;
    ULONG    ulAppRet;
    LONG     ulRc;
    UCHAR    szError[ 256 ],
             szMsgText[ US_STRING_RES_LIMIT + 1 ],
             szMsgTitle[ US_STRING_RES_LIMIT + 1 ];


    switch( msg ) {

        case WM_INITDLG:
            // Save the global data to a window word
            pGlobal = (PATSHGLOBAL) mp2;
            WinSetWindowPtr( hwnd, 0, pGlobal );

            // Set the window mini-icon
            hicon = WinLoadPointer( HWND_DESKTOP, pGlobal->hResources, ID_MAINPROGRAM );
            WinDefDlgProc( hwnd, WM_SETICON, MPFROMP( hicon ), NULL );

            // Load the popup menu
            hwndMenu = WinLoadMenu( hwnd, pGlobal->hResources, ID_MAINPROGRAM );

            // Disable secure controls if requested
            if ( pGlobal->fRunSecure ) {

                WinEnableWindow( WinWindowFromID( hwnd, ID_CONFIGURE ), FALSE );
                WinSendMsg( hwndMenu, MM_SETITEMATTR,
                            MPFROM2SHORT( ID_CONFIGURE, TRUE ), MPFROM2SHORT( MIA_DISABLED, MIA_DISABLED ));
                WinShowWindow( WinWindowFromID( hwnd, ID_CONFIGURE ), FALSE );

                WinEnableWindow( WinWindowFromID( hwnd, ID_STARTUP ), FALSE );
                WinSendMsg( hwndMenu, MM_SETITEMATTR,
                            MPFROM2SHORT( ID_STARTUP, TRUE ), MPFROM2SHORT( MIA_DISABLED, MIA_DISABLED ));

                WinEnableWindow( WinWindowFromID( hwnd, ID_PROMPT ), FALSE );
                WinSendMsg( hwndMenu, MM_SETITEMATTR,
                            MPFROM2SHORT( ID_PROMPT, TRUE ), MPFROM2SHORT( MIA_DISABLED, MIA_DISABLED ));
                WinShowWindow( WinWindowFromID( hwnd, ID_PROMPT ), FALSE );

            }

            // If there's a startup program defined, run it
            if ( pGlobal->pszSProgram != NULL ) {
                LaunchProgram( hwnd, pGlobal->pszSProgram, pGlobal->pszSParams, pGlobal->pszSDir, pGlobal->pszSEnv, TRUE );
                fRunNow = pGlobal->fRunSilent;
                return (MRESULT) FALSE;         // preserve the /q option if specified
            }

            // Now either show the GUI or run the program (as requested)
            if ( pGlobal->fRunSilent )
                LaunchProgram( hwnd, pGlobal->pszProgram, pGlobal->pszParams, pGlobal->pszDir, pGlobal->pszEnv, TRUE );
            else
                WinShowWindow( hwnd, TRUE );

            return (MRESULT) FALSE;


        case WM_COMMAND:
            switch( SHORT1FROMMP( mp1 )) {

                case ID_ABOUT:
                    WinDlgBox( HWND_DESKTOP, hwnd, (PFNWP) AboutDlgProc, pGlobal->hResources, IDD_ABOUT, NULL );
                    break;

                case ID_CONFIGURE:              // Configure the default program
                    if ( ! pGlobal->fRunSecure )
                        DlgConfigure( hwnd, 0 );
                    break;

                case ID_LAUNCH:                 // Launch the default program
                    LaunchProgram( hwnd, pGlobal->pszProgram, pGlobal->pszParams, pGlobal->pszDir, pGlobal->pszEnv, TRUE );
                    break;

                case ID_PROMPT:
                    if ( ! pGlobal->fRunSecure )
                        LaunchPrompt( hwnd );       // Open a CMD.EXE window
                    break;

                case ID_SHUTDOWN:               // Shut down the system
                    ulRc = WinLoadString( pGlobal->hab, pGlobal->hResources,
                                          SZ_SHUTDOWN_CONFIRM, US_STRING_RES_LIMIT, szMsgText );
                    ulRc = WinLoadString( pGlobal->hab, pGlobal->hResources,
                                          SZ_SHUTDOWN_TITLE, US_STRING_RES_LIMIT, szMsgTitle );
                    if ( WinMessageBox( HWND_DESKTOP, hwnd, szMsgText, szMsgTitle,
                                        0, MB_YESNO | MB_QUERY | MB_MOVEABLE ) == MBID_YES )
                    {
                        WinCancelShutdown( pGlobal->hmq, TRUE );
                        fSuccess = WinShutdownSystem( pGlobal->hab, pGlobal->hmq );
                        if ( fSuccess ) WinPostMsg( hwnd, WM_QUIT, 0, 0 );
                    }
                    return (MRESULT) 0;

                case ID_STARTUP:                // Configure the startup program
                    DlgConfigure( hwnd, 1 );
                    break;

                case ID_QUIT:                   // Exit the program
                    WinPostMsg( hwnd, WM_CLOSE, 0, 0 );
                    return (MRESULT) 0;

                default: break;

            } /* end WM_COMMAND messages */
            return (MRESULT) 0;


        case WM_CONTROL:
            switch( SHORT1FROMMP( mp1 )) {
                default: break;
            } /* end WM_CONTROL messages */
            return (MRESULT) 0;


        case WM_CONTEXTMENU:
            if SHORT2FROMMP( mp2 )
                WinQueryPointerPos( HWND_DESKTOP, &mouse );
            else {
                mouse.x = SHORT1FROMMP( mp1 );
                mouse.y = SHORT2FROMMP( mp1 );
                WinMapWindowPoints( hwnd, HWND_DESKTOP, &mouse, 1 );
            }
            WinPopupMenu( HWND_DESKTOP, hwnd, hwndMenu, mouse.x, mouse.y, ID_MAINPROGRAM,
                          PU_HCONSTRAIN | PU_VCONSTRAIN | PU_MOUSEBUTTON1 | PU_MOUSEBUTTON2 | PU_KEYBOARD );
            return 0;


        case WM_APPTERMINATENOTIFY:
            ulAppRet = (ULONG) mp2;
            if ( fRunNow ) {
                // fRunNow is set when fRunSilent is true AND a startup program was executed.
                // It basically preserves the /q option once the startup program exits.
                fRunNow = FALSE;
                LaunchProgram( hwnd, pGlobal->pszProgram, pGlobal->pszParams, pGlobal->pszDir, pGlobal->pszEnv, TRUE );
            } else
                WinShowWindow( hwnd, TRUE );
            break;


        case WM_HELP:
            if (( hwndHelp = WinQueryHelpInstance( hwnd )) != NULLHANDLE )
                WinSendMsg( hwndHelp, HM_GENERAL_HELP, 0, 0 );
            return (MRESULT) 0;


        case WM_CLOSE:
            ulRc = WinLoadString( pGlobal->hab, pGlobal->hResources,
                                  SZ_CLOSE_CONFIRM, US_STRING_RES_LIMIT, szMsgText );
            ulRc = WinLoadString( pGlobal->hab, pGlobal->hResources,
                                  SZ_CLOSE_TITLE, US_STRING_RES_LIMIT, szMsgTitle );
            if ( WinMessageBox( HWND_DESKTOP, hwnd, szMsgText, szMsgTitle,
                                0, MB_OKCANCEL | MB_QUERY | MB_MOVEABLE ) != MBID_CANCEL )
                WinPostMsg( hwnd, WM_QUIT, 0, 0 );
            return (MRESULT) 0;


    } /* end event handlers */

    return WinDefDlgProc( hwnd, msg, mp1, mp2 );
}


/* ------------------------------------------------------------------------- *
 * LocateProfile                                                             *
 *                                                                           *
 * Figure out where to place our INI file.  This will be in the same         *
 * directory as OS2.INI (the OS/2 user profile).                             *
 * ------------------------------------------------------------------------- */
PSZ LocateProfile( void )
{
    ULONG  ulRc;
    USHORT x;
    PSZ    pszOs2Ini,
           pszBuffer,
           pszTok,
           pszProfile;

    // Query the %USER_INI% environment variable which points to OS2.INI
    ulRc      = DosScanEnv("USER_INI", &pszOs2Ini );
    pszBuffer = (PSZ) malloc( strlen(pszOs2Ini) );
    strncpy( pszBuffer, pszOs2Ini, US_PROFILE_LIMIT );

    // Parse the path from the %USER_INI% filespec
    x = strlen( pszBuffer ) - 1;
    while (( x >= 0 ) && ( *(pszBuffer+x) != '\\')) x--;
    pszProfile = (PSZ) malloc( x + strlen(PROFILE) + 1 );
    *(pszBuffer+x) = '\0';

    // Now define our own INI file
    sprintf( pszProfile, "%s\\%s", pszBuffer, PROFILE );

    return ( pszProfile );
}


/* ------------------------------------------------------------------------- *
 * LoadProfileSetting                                                        *
 *                                                                           *
 * Retrieve a value from the program INI file.                               *
 *                                                                           *
 * ARGUMENTS:                                                                *
 *     PSZ *pszData : Referenced address of the buffer to place the data in. *
 *     HINI hIni    : Handle to the INI file.                                *
 *     PSZ  pszApp  : INI application name.                                  *
 *     PSZ  pszKey  : INI key name.                                          *
 *                                                                           *
 * RETURNS: N/A                                                              *
 * ------------------------------------------------------------------------- */
void LoadProfileSetting( PSZ *pszData, HINI hIni, PSZ pszApp, PSZ pszKey )
{
    BOOL  fOK;
    ULONG ulBytes;

    fOK = PrfQueryProfileSize( hIni, pszApp, pszKey, &ulBytes );
    if ( fOK && ( ulBytes > 0 )) {
        if ( *pszData != NULL ) free( *pszData );     // to avoid memory leak
        *pszData = (PSZ) malloc( ulBytes );
        PrfQueryProfileData( hIni, pszApp, pszKey, *pszData, &ulBytes );
        if ( strcmp( *pszData, "\0") == 0 ) {
            free( *pszData );
            *pszData = NULL;
        }
    }
}


/* ------------------------------------------------------------------------- *
 * CentreWindow                                                              *
 *                                                                           *
 * Centres the given window on the screen.                                   *
 *                                                                           *
 * ARGUMENTS:                                                                *
 *     HWND hwnd: handle of the window to be centred.                        *
 *                                                                           *
 * RETURNS: N/A                                                              *
 * ------------------------------------------------------------------------- */
void CentreWindow( HWND hwnd )
{
    LONG scr_width, scr_height,
         app_width, app_height;
    LONG x, y;
    SWP wp;

    scr_width = WinQuerySysValue( HWND_DESKTOP, SV_CXSCREEN );
    scr_height = WinQuerySysValue( HWND_DESKTOP, SV_CYSCREEN );

    if ( WinQueryWindowPos( hwnd, &wp )) {
        x = ( scr_width - wp.cx ) / 2;
        y = ( scr_height - wp.cy ) / 2;
        WinSetWindowPos( hwnd, HWND_TOP, x, y, wp.cx, wp.cy, SWP_MOVE | SWP_ACTIVATE );
    }

}


/* ------------------------------------------------------------------------- *
 * LaunchPrompt                                                              *
 *                                                                           *
 * Launches a windowable CMD.EXE session, with the working directory set to  *
 * the root of the boot drive.                                               *
 *                                                                           *
 * ARGUMENTS:                                                                *
 *     HWND hwnd: handle of the main application window.                     *
 *                                                                           *
 * RETURNS: N/A                                                              *
 * ------------------------------------------------------------------------- */
void LaunchPrompt( HWND hwnd )
{
    ULONG ulRc,
          ulOrd;
    UCHAR szBootDrv[ 4 ];


    ulRc = DosQuerySysInfo( QSV_BOOT_DRIVE, QSV_BOOT_DRIVE, &ulOrd, sizeof(ULONG) );
    if (( ulRc != NO_ERROR ) || ( ulOrd > 26 ) || ( ulOrd < 1 ))
        LaunchProgram( hwnd, "CMD.EXE", NULL, NULL, NULL, FALSE );
    else {
        sprintf( szBootDrv, "%c:\\", 64 + ulOrd );
        LaunchProgram( hwnd, "CMD.EXE", NULL, szBootDrv, NULL, FALSE );
    }
}


/* ------------------------------------------------------------------------- *
 * LaunchProgram                                                             *
 *                                                                           *
 * Launches the selected program.                                            *
 *                                                                           *
 * ARGUMENTS:                                                                *
 *     HWND hwnd      : handle of the main application window.               *
 *     PSZ pszProgram : program path and filename.                           *
 *     PSZ pszParams  : program parameters.                                  *
 *     PSZ pszDir     : program startup directory.                           *
 *     PSZ pszEnv     : string to append to program environment.             *
 *     BOOL fHide     : determines if ATSHELL should hide when program runs. *
 *                                                                           *
 * RETURNS: N/A                                                              *
 * ------------------------------------------------------------------------- */
void LaunchProgram( HWND hwnd, PSZ pszProgram, PSZ pszParams, PSZ pszDir, PSZ pszEnv, BOOL fHide )
{
    PATSHGLOBAL  pGlobal;
    PPROGDETAILS pDetails;
    HAPP         happ;
    UCHAR        szError[ US_STRING_RES_LIMIT + 20 ],
                 szMsgText[ US_STRING_RES_LIMIT + 1 ],
                 szMsgTitle[ US_STRING_RES_LIMIT + 1 ];
    ULONG        ulErr;
    LONG         ulRc;


    // TODO: get the current environment and append pszEnv.

    pGlobal = WinQueryWindowPtr( hwnd, 0 );

    pDetails                              = (PPROGDETAILS) malloc( sizeof(PROGDETAILS) );
    pDetails->Length                      = sizeof( PROGDETAILS );
    pDetails->progt.progc                 = PROG_DEFAULT;
    pDetails->progt.fbVisible             = SHE_VISIBLE;
    pDetails->pszTitle                    = "";
    pDetails->pszExecutable               = pszProgram;
    pDetails->pszParameters               = pszParams;
    pDetails->pszStartupDir               = pszDir;
    pDetails->pszIcon                     = NULL;
    pDetails->pszEnvironment              = pszEnv;
    pDetails->swpInitial.fl               = SWP_ACTIVATE | SWP_ZORDER | SWP_SHOW;
    pDetails->swpInitial.cy               = 0;
    pDetails->swpInitial.cx               = 0;
    pDetails->swpInitial.y                = 0;
    pDetails->swpInitial.x                = 0;
    pDetails->swpInitial.hwndInsertBehind = HWND_TOP;
    pDetails->swpInitial.hwnd             = hwnd;
    pDetails->swpInitial.ulReserved1      = 0;
    pDetails->swpInitial.ulReserved2      = 0;

    happ = WinStartApp( hwnd, pDetails, NULL, NULL, SAF_INSTALLEDCMDLINE | SAF_STARTCHILDAPP );
    if ( happ == NULLHANDLE ) {
        ulErr = WinGetLastError( pGlobal->hab );
        ulRc = WinLoadString( pGlobal->hab, pGlobal->hResources,
                              SZ_EXECUTE_ERROR, US_STRING_RES_LIMIT, szMsgText );
        ulRc = WinLoadString( pGlobal->hab, pGlobal->hResources,
                              SZ_EXECUTE_TITLE, US_STRING_RES_LIMIT, szMsgTitle );
        sprintf( szError, szMsgText, ERRORIDSEV( ulErr ), ERRORIDERROR( ulErr ));
        WinMessageBox( HWND_DESKTOP, HWND_DESKTOP, szError, szMsgTitle, 0, MB_CANCEL | MB_ERROR );
        WinShowWindow( hwnd, TRUE );
    }
    else if ( fHide) WinShowWindow( hwnd, FALSE );

    free( pDetails );
}


/* ------------------------------------------------------------------------- *
 * DlgConfigure                                                              *
 *                                                                           *
 * Opens the program configuration dialog.                                   *
 *                                                                           *
 * ARGUMENTS:                                                                *
 *     HWND hwnd        : handle of the main application window.             *
 *     USHORT usProgram : identifies which program will be configured.       *
 *                        0 = default program; 1 = startup program           *
 *                                                                           *
 * RETURNS: N/A                                                              *
 * ------------------------------------------------------------------------- */
void DlgConfigure( HWND hwnd, USHORT usProgram )
{
    PATSHGLOBAL    pGlobal;
    ATSHAPPCONFIG  dlgOptions;
    ULONG  ulRc,
           ulOrd;


    pGlobal = WinQueryWindowPtr( hwnd, 0 );

    dlgOptions.hab        = pGlobal->hab;
    dlgOptions.hResources = pGlobal->hResources;
    dlgOptions.fChanged   = FALSE;
    ulRc = DosQuerySysInfo( QSV_MAX_PATH_LENGTH, QSV_MAX_PATH_LENGTH, &ulOrd, sizeof(ULONG) );
    if (( ulRc == NO_ERROR ) && ( ulOrd > 12 )) dlgOptions.ulPathLimit = ulOrd;
    else                                        dlgOptions.ulPathLimit = US_DEF_FILESPEC_LIMIT + 1;

    if ( usProgram > 0 ) {
        dlgOptions.pszProgram = pGlobal->pszSProgram;
        dlgOptions.pszParams  = pGlobal->pszSParams;
        dlgOptions.pszDir     = pGlobal->pszSDir;
        dlgOptions.pszEnv     = pGlobal->pszSEnv;
    } else {
        dlgOptions.pszProgram = pGlobal->pszProgram;
        dlgOptions.pszParams  = pGlobal->pszParams;
        dlgOptions.pszDir     = pGlobal->pszDir;
        dlgOptions.pszEnv     = pGlobal->pszEnv;
    }

    WinDlgBox( HWND_DESKTOP, hwnd, (PFNWP) AppCfgDlgProc,
               pGlobal->hResources, IDD_PROGCONFIG, &dlgOptions );

    if ( dlgOptions.fChanged ) {
        if ( usProgram > 0 ) {
            if ( pGlobal->pszSProgram != NULL ) free ( pGlobal->pszSProgram );
            if ( pGlobal->pszSParams  != NULL ) free ( pGlobal->pszSParams  );
            if ( pGlobal->pszSDir     != NULL ) free ( pGlobal->pszSDir     );
            if ( pGlobal->pszSEnv     != NULL ) free ( pGlobal->pszSEnv     );
            pGlobal->pszSProgram = dlgOptions.pszProgram;
            pGlobal->pszSParams  = dlgOptions.pszParams;
            pGlobal->pszSDir     = dlgOptions.pszDir;
            pGlobal->pszSEnv     = dlgOptions.pszEnv;
            // Save settings to profile
            PrfWriteProfileString( pGlobal->hIni, PRF_STARTUP_PROGRAM,
                                   PRF_KEY_PROGRAM, pGlobal->pszSProgram );
            PrfWriteProfileString( pGlobal->hIni, PRF_STARTUP_PROGRAM,
                                   PRF_KEY_PARAMS, pGlobal->pszSParams   );
            PrfWriteProfileString( pGlobal->hIni, PRF_STARTUP_PROGRAM,
                                   PRF_KEY_DIR, pGlobal->pszSDir         );
            PrfWriteProfileString( pGlobal->hIni, PRF_STARTUP_PROGRAM,
                                   PRF_KEY_ENV, pGlobal->pszSEnv         );
        } else {
            if ( pGlobal->pszSProgram != NULL ) free ( pGlobal->pszProgram );
            if ( pGlobal->pszSParams  != NULL ) free ( pGlobal->pszParams  );
            if ( pGlobal->pszSDir     != NULL ) free ( pGlobal->pszDir     );
            if ( pGlobal->pszSEnv     != NULL ) free ( pGlobal->pszEnv     );
            pGlobal->pszProgram = dlgOptions.pszProgram;
            pGlobal->pszParams  = dlgOptions.pszParams;
            pGlobal->pszDir     = dlgOptions.pszDir;
            pGlobal->pszEnv     = dlgOptions.pszEnv;
            // Save settings to profile
            PrfWriteProfileString( pGlobal->hIni, PRF_DEFAULT_PROGRAM,
                                   PRF_KEY_PROGRAM, pGlobal->pszProgram );
            PrfWriteProfileString( pGlobal->hIni, PRF_DEFAULT_PROGRAM,
                                   PRF_KEY_PARAMS, pGlobal->pszParams   );
            PrfWriteProfileString( pGlobal->hIni, PRF_DEFAULT_PROGRAM,
                                   PRF_KEY_DIR, pGlobal->pszDir         );
            PrfWriteProfileString( pGlobal->hIni, PRF_DEFAULT_PROGRAM,
                                   PRF_KEY_ENV, pGlobal->pszEnv         );
        }
    }

}


/* ------------------------------------------------------------------------- *
 * AppCfgDlgProc                                                             *
 *                                                                           *
 * Dialog procedure for the program configuration dialog.                    *
 * ------------------------------------------------------------------------- */
MRESULT EXPENTRY AppCfgDlgProc( HWND hwnd, ULONG msg, MPARAM mp1, MPARAM mp2 )
{
    static PATSHAPPCONFIG pOptions;
    PSZ    pszProgram,
           pszParams,
           pszDir,
           pszEnv;
    ULONG  ulItem;


    switch ( msg ) {

        case WM_INITDLG:
            pOptions = (PATSHAPPCONFIG) mp2;
            CentreWindow( hwnd );
            WinSendDlgItemMsg( hwnd, IDD_PROGNAME,  EM_SETTEXTLIMIT, MPFROMLONG( pOptions->ulPathLimit ),  0 );
            WinSendDlgItemMsg( hwnd, IDD_PROGPARMS, EM_SETTEXTLIMIT, MPFROMSHORT( US_PARAMETERS_LIMIT ),   0 );
            WinSendDlgItemMsg( hwnd, IDD_PROGDIR,   EM_SETTEXTLIMIT, MPFROMLONG( pOptions->ulPathLimit ),  0 );
            WinSendDlgItemMsg( hwnd, IDD_PROGENV,   EM_SETTEXTLIMIT, MPFROMSHORT( US_ENVIRONMENT_LIMIT ),  0 );
            WinSetDlgItemText( hwnd, IDD_PROGNAME,  pOptions->pszProgram );
            WinSetDlgItemText( hwnd, IDD_PROGPARMS, pOptions->pszParams );
            WinSetDlgItemText( hwnd, IDD_PROGDIR,   pOptions->pszDir );
            // WinSetDlgItemText( hwnd, IDD_PROGENV,   pOptions->pszEnv );
            break;

        case WM_COMMAND:
            switch ( SHORT1FROMMP( mp1 )) {
                case IDD_PROGBROWSE:
                    BrowseProgram( hwnd );
                    return (MRESULT) 0;

                case DID_OK:
                    // Query program filespec
                    pszProgram = (PSZ) malloc( pOptions->ulPathLimit + 1 );
                    if ( strcmp( pszProgram, "") == 0 ) pszProgram = NULL;
                    WinQueryDlgItemText( hwnd, IDD_PROGNAME, pOptions->ulPathLimit, pszProgram );
                    if (( pOptions->pszProgram == NULL ) || ( strncmp( pOptions->pszProgram, pszProgram, pOptions->ulPathLimit ) != 0 )) {
                        pOptions->pszProgram = pszProgram;
                        pOptions->fChanged   = TRUE;
                    }
                    // Query parameters
                    pszParams = (PSZ) malloc( US_PARAMETERS_LIMIT + 1 );
                    if ( strcmp( pszParams, "") == 0 ) pszParams = NULL;
                    WinQueryDlgItemText( hwnd, IDD_PROGPARMS, US_PARAMETERS_LIMIT, pszParams );
                    if (( pOptions->pszParams == NULL ) || ( strncmp( pOptions->pszParams, pszParams, US_PARAMETERS_LIMIT ) != 0 )) {
                        pOptions->pszParams = pszParams;
                        pOptions->fChanged  = TRUE;
                    }
                    // Query startup directory
                    pszDir = (PSZ) malloc( pOptions->ulPathLimit + 1 );
                    if ( strcmp( pszDir, "") == 0 ) pszDir = NULL;
                    WinQueryDlgItemText( hwnd, IDD_PROGDIR, pOptions->ulPathLimit, pszDir );
                    if (( pOptions->pszDir == NULL) || ( strncmp( pOptions->pszDir, pszDir, pOptions->ulPathLimit ) != 0 )) {
                        pOptions->pszDir   = pszDir;
                        pOptions->fChanged = TRUE;
                    }
                    // TODO: Query environment variable additions from combo-box and concatenate them together with \0.
                    break;

                case IDD_PROGENVADD:
                    pszEnv = (PSZ) malloc( US_ENVIRONMENT_LIMIT + 1 );
                    WinQueryDlgItemText( hwnd, IDD_PROGENV, US_ENVIRONMENT_LIMIT, pszEnv );
                    WinSendDlgItemMsg( hwnd, IDD_PROGENV, LM_INSERTITEM, MPFROMSHORT( LIT_END ), MPFROMP( pszEnv ));
                    return (MRESULT) 0;

                case IDD_PROGENVDEL:
                    ulItem = (ULONG) WinSendDlgItemMsg( hwnd, IDD_PROGENV, LM_QUERYSELECTION, MPFROMSHORT( LIT_CURSOR ), 0 );
                    if ( ulItem != LIT_NONE )
                        WinSendDlgItemMsg( hwnd, IDD_PROGENV, LM_DELETEITEM, MPFROMSHORT( ulItem ), 0 );
                    return (MRESULT) 0;

                default: break;
            }
            break;

        default: break;
    }

    return WinDefDlgProc( hwnd, msg, mp1, mp2 );
}


/* ------------------------------------------------------------------------- *
 * BrowseProgram                                                             *
 *                                                                           *
 * Opens a File dialog for the user to select a program after clicking on    *
 * 'Browse' from the program configuration dialog.                           *
 *                                                                           *
 * ARGUMENTS:                                                                *
 *     HWND hwnd: handle of the main application window.                     *
 *                                                                           *
 * RETURNS: N/A                                                              *
 * ------------------------------------------------------------------------- */
void BrowseProgram( HWND hwnd )
{
    FILEDLG fild;
    HWND hwndDlg;


    memset( &fild, 0, sizeof(FILEDLG) );
    fild.cbSize   = sizeof( FILEDLG );
    fild.fl       = FDS_HELPBUTTON | FDS_CENTER | FDS_OPEN_DIALOG;
    fild.pszTitle = "Select Program";
    strcpy( fild.szFullFile, "*.EXE");

    hwndDlg = WinFileDlg( HWND_DESKTOP, hwnd, &fild );
    if ( hwndDlg && (fild.lReturn == DID_OK ))
    {
        WinSetDlgItemText( hwnd, IDD_PROGNAME, fild.szFullFile );
    }

}


/* ------------------------------------------------------------------------- *
 * AboutDlgProc                                                              *
 *                                                                           *
 * Dialog procedure for the product information dialog.                      *
 * ------------------------------------------------------------------------- */
MRESULT EXPENTRY AboutDlgProc( HWND hwnd, ULONG msg, MPARAM mp1, MPARAM mp2 )
{
    switch ( msg ) {

        case WM_INITDLG:
            CentreWindow( hwnd );
            break;

        default: break;
    }

    return WinDefDlgProc( hwnd, msg, mp1, mp2 );
}

